import sys
if sys.platform == "pyboard":
    import pyb
class Editor:
    KEYMAP = { 
    b"\x1b[A" : 0x4001,
    b"\x1b[B" : 0x4002,
    b"\x1b[D" : 0x4003,
    b"\x1b[C" : 0x4004,
    b"\x0b" : 0x4001, 
    b"\x0a" : 0x4002, 
    b"\x08" : 0x4003, 
    b"\x0c" : 0x4004,
    b"\x1b[H" : 0x4005, 
    b"\x1bOH" : 0x4005, 
    b"\x1b[1~": 0x4005, 
    b"\x17" : 0x4005, 
    b"\x1b[F" : 0x4006, 
    b"\x1bOF" : 0x4006, 
    b"\x1b[4~": 0x4006, 
    b"\x05" : 0x4006, 
    b"\x1b[5~": 0x4007,
    b"\x0f" : 0x4007, 
    b"\x1b[6~": 0x4008,
    b"\x10" : 0x4008, 
    b"\x11" : 0x4009, 
    b"\r" : 0x400a,
    b"\x7f" : 0x400b, 
    b"\x1b[3~": 0x400c,
    b"\x19" : 0x400c, 
    b"\x13" : 0x400d, 
    b"\x06" : 0x4010, 
    b"\x0e" : 0x4014, 
    b"\x07" : 0x4011, 
    b"\x14" : 0x4012, 
    b"\x1b[1;5H": 0x4012,
    b"\x02" : 0x4013, 
    b"\x1b[1;5F": 0x4013,
    b"\x1b[3;5~": 0x4015,
    b"\x18" : 0x4015, 
    b"\x09" : 0x400e,
    b"\x1b[Z" : 0x400f, 
    b"\x15" : 0x400f, 
    b"\x16" : 0x4017, 
    b"\x1bOP" : 0x4016, 
    b"\x1bOQ" : 0x4016, 
    b"\x01" : 0x4018, 
    }
    def __init__(self, width, height, tab_size, status):
        self.top_line = 0
        self.cur_line = 0
        self.row = 0
        self.col = 0
        self.margin = 0
        self.height = height - 1
        self.width = width
        self.hstep = int(width / 6)
        self.k_buffer = b""
        self.tab_size = tab_size
        self.status = status
        self.autoindent = True
        self.changed = ' '
        self.message = ""
        self.find_pattern = ""
        self.y_buffer = []
        self.lastkey = 0
    if sys.platform == "pyboard":
        @staticmethod
        def wr(s):
            ns = 0
            while ns < len(s): 
                res = Editor.serialcomm.write(s[ns:])
                if res != None:
                    ns += res
        @staticmethod
        def rd():
            while not Editor.serialcomm.any():
                pass
            return Editor.serialcomm.read(1)
    @staticmethod
    def cls():
        Editor.wr(b"\x1b[2J")
    @staticmethod
    def goto(row, col):
        
        Editor.wr("\x1b[%d;%dH" % (row + 1, col + 1))
    @staticmethod
    def clear_to_eol():
        Editor.wr(b"\x1b[0K")
    @staticmethod
    def cursor(onoff):
        if onoff:
            Editor.wr(b"\x1b[?25h")
        else:
            Editor.wr(b"\x1b[?25l")
    @staticmethod
    def hilite(onoff):
        if onoff:
            Editor.wr(b"\x1b[1m")
        else:
            Editor.wr(b"\x1b[0m")
    def c_or_f(self): 
        if len(self.k_buffer) == 0:
            self.k_buffer += Editor.rd() 
        while True:
            for k in self.KEYMAP.keys():
                if self.k_buffer == k[:len(self.k_buffer)]: 
                    if self.k_buffer == k:
                        c = self.KEYMAP[self.k_buffer]
                        self.k_buffer = b""
                        return c 
                    else: 
                        break
            else: 
                c = self.k_buffer[0]
                self.k_buffer = self.k_buffer[1:]
                if c >= ord(' '): 
                    return c
            self.k_buffer += Editor.rd() 
    def adjust_cursor_eol(self, updt): 
        self.col = min(self.col, len(self.content[self.cur_line]) - self.margin)
        if not (0 <= self.col < self.width): 
            self.goto_col(True)
    def goto_col(self, updt):
        
            if self.col >= self.width:
                self.margin = self.col + self.margin - (self.width - 1) + self.hstep
                self.col = self.width - 1 - self.hstep
                self.update_screen()
            elif self.col < 0:
                val = self.col + self.margin 
                self.margin = max(self.margin - self.width, 0)
                self.col = val - self.margin
                self.update_screen()
            else:
                if updt: self.update_line()
    
    
    def goto_line(self, no):
        
        self.cur_line = no
        if self.top_line <= no < self.top_line + self.height: 
            self.row = no - self.top_line
            self.adjust_cursor_eol(True) 
        else:
            self.top_line = no - self.row
            if self.top_line < 0:
                self.top_line = 0
                self.row = no
            self.adjust_cursor_eol(False) 
            self.update_screen() 
    def set_lines(self, lines, fname):
        self.content = lines
        self.total_lines = len(lines)
        self.fname = fname
    def update_screen(self):
        self.cursor(False)
        self.cls()
        self.cursor(False)
        i = self.top_line
        for c in range(self.height):
            self.goto(c, 0)
            if i == self.total_lines:
                self.clear_to_eol()
            else:
                self.show_line(self.content[i])
                i += 1
        self.cursor(True)
    def update_line(self):
        self.cursor(False)
        self.goto(self.row, 0)
        self.show_line(self.content[self.cur_line])
        self.cursor(True)
    def show_line(self, l):
        l = l[self.margin:]
        l = l[:self.width]
        self.wr(l)
        if len(l) < self.width: self.clear_to_eol()
    def show_status(self):
        if self.status or self.message:
            self.cursor(False)
            self.goto(self.height, 0)
            self.hilite(True)
            self.wr("%c Ln: %d Col: %d  %s %d" %
                    (self.changed, self.cur_line + 1, self.col + self.margin + 1, self.message, self.margin))
            self.clear_to_eol()
            self.hilite(False)
            self.cursor(True)
    def clear_status(self):
        if (not self.status) and self.message:
            self.goto(self.height, 0)
            self.clear_to_eol()
        self.message = ''
    def next_line(self):
        if self.row + 1 == self.height:
            self.top_line += 1
            return True
        else:
            self.row += 1
            return False
    def spaces(self, line, pos = 0): 
        if pos: 
            return len(line[:pos]) - len(line[:pos].rstrip(" "))
        else: 
            return len(line) - len(line.lstrip(" "))
    def line_edit(self, prompt, default): 
        self.goto(self.height, 0)
        self.hilite(True)
        self.wr(prompt)
        self.wr(default)
        self.clear_to_eol()
        res = default
        self.message = ' ' 
        while True:
            key = self.c_or_f() 
            if key == 0x400a: 
                self.hilite(False)
                return res
            elif key == 0x4009: 
                self.hilite(False)
                return None
            elif key in (0x400b, 0x400c): 
                if (len(res) > 0):
                    res = res[:len(res)-1]
                    self.wr('\b \b')
            elif 0x20 <= key < 0x100: 
                res += chr(key)
                self.wr(chr(key))
            else: 
                pass
    def find_in_file(self, pattern, pos):
        self.find_pattern = pattern 
        spos = pos + self.margin
        for line in range(self.cur_line, self.total_lines):
            match = self.content[line][spos:].lower().find(pattern)
            if match >= 0:
                break
            spos = 0
        else:
            self.message = pattern + " not found"
            return True
        self.col = match - self.margin
        self.goto_col(False)
        self.goto_line(line)
    def handle_cursor_keys(self, key): 
        if key == 0x4002:
            if self.cur_line + 1 < self.total_lines:
                self.cur_line += 1
                self.goto_line(self.cur_line)
        elif key == 0x4001:
            if self.cur_line > 0:
                self.cur_line -= 1
                self.goto_line(self.cur_line)
        elif key == 0x4003:
            self.col -= 1
            self.goto_col(False)
        elif key == 0x4004:
            self.col += 1
            self.adjust_cursor_eol(True)
        elif key == 0x4005:
            ns = self.spaces(self.content[self.cur_line])
            if self.col + self.margin > ns:
                self.col = ns - self.margin
            else:
                self.col = -self.margin
            self.goto_col(False)
        elif key == 0x4006:
            self.col = len(self.content[self.cur_line]) - self.margin
            self.adjust_cursor_eol(True)
        elif key == 0x4007:
            self.cur_line -= self.height
            if self.cur_line < 0:
                self.cur_line = 0
            self.goto_line(self.cur_line)
        elif key == 0x4008:
            self.cur_line += self.height
            if self.cur_line >= self.total_lines:
                self.cur_line = self.total_lines - 1
            self.goto_line(self.cur_line)
        elif key == 0x4010:
            pat = self.line_edit("Find: ", self.find_pattern)
            if pat:
                self.find_in_file(pat.lower(), self.col)
        elif key == 0x4014:
            self.find_in_file(self.find_pattern, self.col + 1)
            self.message = ' ' 
        elif key == 0x4011: 
            line = self.line_edit("Goto Line: ", "")
            if line:
                line = min(self.total_lines - 1, max(int(line) - 1, 0))
                self.goto_line(line)
        elif key == 0x4012: 
            self.goto_line(0)
        elif key == 0x4013: 
            self.goto_line(self.total_lines - 1)
            self.message = ' ' 
        elif key == 0x4018: 
            self.autoindent = not self.autoindent
            self.message = "Autoindent %s" % self.autoindent
        else:
            return False
        return True
    def handle_buffer_keys(self, key): 
            if key == 0x400d:
                fname = self.line_edit("File Name: ", self.fname)
                if fname:
                    try:
                        with open(fname, "w") as f:
                            self.wr(" ..Saving..")
                            for l in self.content:
                                f.write(l + '\n')
                        f.close()
                        self.changed = " "
                    except:
                        pass
            elif key == 0x4016:
                try:
                    with open("pybe.hlp") as f:
                        self.cls()
                        ln = 0
                        for l in f:
                            self.goto(ln, 0)
                            self.wr(l.strip("\r\n")[:self.width])
                            ln += 1
                            if ln > self.height: break
                    f.close()
                    self.line_edit("Press Enter to continue", "")
                    self.update_screen()
                except:
                    self.message = "Help file not found!"
            else:
                return False
            return True
    def handle_key(self, key): 
            l = self.content[self.cur_line]
            sc = self.changed
            self.changed = '*'
            if key == 0x400a:
                self.content[self.cur_line] = l[:self.col + self.margin]
                if self.autoindent and self.col + self.margin > 0:
                    ni = self.spaces(l, 0)
                else:
                    ni = 0
                self.cur_line += 1
                self.content[self.cur_line:self.cur_line] = [' ' * ni + l[self.col + self.margin:]]
                self.total_lines += 1
                self.col = ni
                self.margin = 0
                self.next_line()
                self.update_screen()
            elif key == 0x400b:
                if self.col + self.margin:
                    self.content[self.cur_line] = l[:self.col + self.margin - 1] + l[self.col + self.margin:]
                    self.col -= 1
                    self.goto_col(True)
                elif self.cur_line: 
                    self.col = len(self.content[self.cur_line - 1])
                    self.content[self.cur_line - 1] += l
                    del self.content[self.cur_line]
                    self.cur_line -= 1
                    self.total_lines -= 1
                    self.goto_col(False)
                    self.goto_line(self.cur_line)
                    self.update_screen()
                else:
                    self.changed = sc
            elif key == 0x400c:
                if (self.col + self.margin) < len(l):
                    l = l[:self.col + self.margin] + l[self.col + self.margin + 1:]
                    self.content[self.cur_line] = l
                    self.update_line()
                elif (self.cur_line + 1) < self.total_lines: 
                    self.content[self.cur_line] = l + self.content.pop(self.cur_line + 1)
                    self.total_lines -= 1
                    self.update_screen()
                else:
                    self.changed = sc
            elif key == 0x4015: 
                if key == self.lastkey: 
                    self.y_buffer.append(l) 
                else:
                    self.y_buffer.clear() 
                    self.y_buffer = [l]
                if self.total_lines > 1: 
                    del self.content[self.cur_line]
                    self.total_lines -= 1
                    if self.cur_line >= self.total_lines: 
                        self.cur_line -= 1
                else: 
                    self.content[self.cur_line] = ''
                self.goto_line(self.cur_line)
                self.update_screen()
            elif key == 0x400e: 
                ns = self.spaces(l, 0)
                ni = self.tab_size - ns % self.tab_size
                self.content[self.cur_line] = l[:self.col + self.margin] + ' ' * ni + l[self.col + self.margin:]
                if ns == len(l) or self.col + self.margin >= ns: 
                    self.col += ni 
                self.goto_col(True)
            elif key == 0x400f: 
                ns = self.spaces(l, 0)
                if ns and self.col + self.margin < ns: 
                    ni = (ns - 1) % self.tab_size + 1
                    self.content[self.cur_line] = l[ni:]
                    self.goto_col(True)
                else: 
                    ns = self.spaces(l, self.col + self.margin)
                    ni = (self.col + self.margin - 1) % self.tab_size + 1
                    if (ns >= ni):
                        self.content[self.cur_line] = l[:self.col + self.margin - ni] + l[self.col + self.margin:]
                        self.col -= ni
                        self.goto_col(True)
                    else:
                        self.changed = sc
            elif key == 0x4017: 
                if self.y_buffer:
                    self.content[self.cur_line:self.cur_line] = self.y_buffer 
                    self.total_lines += len(self.y_buffer)
                    self.update_screen()
                else:
                    self.changed = sc
            elif 32 <= key < 0x4000:
                self.content[self.cur_line] = l[:self.col + self.margin] + chr(key) + l[self.col + self.margin:]
                self.col += 1
                self.goto_col(True)
            else: 
                self.changed = sc
                pass
    def loop(self): 
        self.update_screen()
        while True:
            self.show_status()
            self.goto(self.row, self.col) 
            key = self.c_or_f() 
            self.clear_status()
            if key == 0x4009:
                if self.changed != ' ':
                    res = self.line_edit("Content changed! Quit without saving (y/N)? ", "N")
                    if not res or res[0].upper() != 'Y':
                        continue
                return None
            elif self.handle_buffer_keys(key):
                pass
            elif self.handle_cursor_keys(key):
                pass
            else: self.handle_key(key)
            self.lastkey = key
    def init_tty(self, device, baud):
        if sys.platform == "pyboard":
            if (device):
                Editor.serialcomm = pyb.UART(device, baud)
            else:
                Editor.serialcomm = pyb.USB_VCP()
        Editor.sdev = device
        pass
    def deinit_tty(self):
        
        self.goto(self.height, 0)
        self.clear_to_eol()
def pybe(name="", content=[""], width=80, height=24, tab_size=4, status=True, device=0, baud=38400):
    if name:
        f = open(name)
        content = []
        for l in f:
            content.append(l.strip('\r\n'))
        f.close()
    e = Editor(width, height, tab_size, status)
    e.init_tty(device, baud)
    e.set_lines(content, name)
    e.loop()
    e.deinit_tty()
 
    if name:
        content.clear()
    if sys.platform == "pyboard":
        import gc
        gc.collect()
